@@ -33,6 +33,8 @@
 
 #include <libart_lgpl/art_affine.h>
 #include <libgnomeprint/gpa/gpa-node.h>
+#include <libgnomeprint/gpa/gpa-node-private.h>
+#include <libgnomeprint/gpa/gpa-key.h>
 #include <libgnomeprint/gnome-print-private.h>
 #include <libgnomeprint/gnome-print-meta.h>
 #include <libgnomeprint/gnome-print-transport.h>
@@ -49,7 +51,7 @@ struct _GnomePrintJob {
 
 	GnomePrintConfig *config;
 	GnomePrintContext *meta;
-	guchar *input_file;
+	gchar *input_file;
 
 	GnomePrintJobPrivate *priv;
 };
@@ -160,8 +162,6 @@ gnome_print_job_init (GnomePrintJob *job)
 
 	job->meta = g_object_new (GNOME_TYPE_PRINT_META, NULL);
 
-	job->input_file = NULL;
-
 	job->priv = g_new0 (GnomePrintJobPrivate, 1);
 	job_clear_config_data (job);
 }
@@ -316,46 +316,89 @@ gnome_print_job_get_context (GnomePrintJob *job)
 void
 gnome_print_job_set_file (GnomePrintJob *job, gchar *input_file)
 {
-	g_return_if_fail (job != NULL);
 	g_return_if_fail (GNOME_IS_PRINT_JOB (job));
-	g_return_if_fail (input_file != NULL);
-	g_return_if_fail (job->input_file == NULL);
 
-	job->input_file = (guchar *) g_strdup (input_file);
+	if (job->input_file) {
+		g_free (job->input_file);
+		job->input_file = NULL;
+		gnome_print_config_set (job->config,
+			(const guchar *) "Settings.Document.Source",
+			(const guchar *) "");
+		
+	}
+	if (input_file) {
+		GPANode *node;
+
+		job->input_file = g_strdup (input_file);
+		node = gpa_node_lookup (GNOME_PRINT_CONFIG_NODE (job->config),
+			(const guchar *) "Settings.Document.Source");
+		if (!node) {
+			node = gpa_node_lookup (GNOME_PRINT_CONFIG_NODE (job->config),
+				(const guchar *) "Settings.Document");
+			gpa_key_insert (node, (const guchar *) "Source",
+				(const guchar *) input_file);
+		}
+		gnome_print_config_set (job->config,
+			(const guchar *) "Settings.Document.Source",
+			(const guchar *) input_file);
+	}
 }
 
 static void
 gnome_print_job_setup_context (GnomePrintJob *job, GnomePrintContext *pc)
 {
-	GnomePrintFilter *filter = NULL;
-	gchar *description;
+	GnomePrintFilter *filter = NULL, *f = NULL;
+	gchar *d;
 
 	g_return_if_fail (GNOME_IS_PRINT_JOB (job));
 	g_return_if_fail (GNOME_IS_PRINT_CONTEXT (pc));
 
 	g_object_set (G_OBJECT (pc), "filter", NULL, NULL);
 
-	/* Explicit filter description */
-	description = (gchar *) gnome_print_config_get (job->config,
+	/*
+	 * There are three places for filters:
+	 * (1) Filters specific to printers (like GnomePrintFilterFrgba) are
+	 *     stored in Settings.Output.Job.Filter.
+	 * (2) Filters independent from printers (like GnomePrintFilterSelect) are
+	 *     stored in Settings.Document.Filter.
+	 * (3) A multipage filter may be specified in the layout key in 
+	 *     the print configuration. We continue to load it, but this 
+	 *     key should be removed (it is not as flexible as a filter
+	 *     pipeline).
+	 * We load the second one first and append the printer specific filter.
+	 */
+	d = (gchar *) gnome_print_config_get (job->config,
 			(const guchar *) "Settings.Output.Job.Filter");
-	if (description) {
-		GError *e = NULL;
-
-		filter = gnome_print_filter_new_from_description (description, &e);
-		if (e) {
-			g_warning ("Problem while creating filter from '%s': %s",
-				description, e->message);
-			g_error_free (e);
-		}
-		g_free (description);
+	if (d) {
+		filter = gnome_print_filter_new_from_description (d, NULL);
+		g_free (d);
 		if (filter) {
-			g_object_set (G_OBJECT (pc), "filter", filter, NULL);
+			f = g_object_new (GNOME_TYPE_PRINT_FILTER, NULL);
+			gnome_print_filter_add_filter (f, filter);
 			g_object_unref (G_OBJECT (filter));
+			g_object_set (G_OBJECT (pc), "filter", f, NULL);
+			g_object_unref (G_OBJECT (f));
+		}
+	}
+	d = (gchar *) gnome_print_config_get (job->config,
+			(const guchar *) "Settings.Document.Filter");
+	if (d) {
+		filter = gnome_print_filter_new_from_description (d, NULL);
+		g_free (d);
+		if (filter) {
+			f = g_object_new (GNOME_TYPE_PRINT_FILTER, NULL);
+			gnome_print_filter_add_filter (f, filter);
+			g_object_unref (G_OBJECT (filter));
+			g_object_get (G_OBJECT (pc), "filter", &filter, NULL);
+			if (filter)
+				gnome_print_filter_append_predecessor (filter, f);
+			g_object_set (G_OBJECT (pc), "filter", f, NULL);
+			g_object_unref (G_OBJECT (f));
 		}
 	}
 
 	gnome_print_job_update_layout_data (job);
-	if (job->priv->LY_LIST && g_list_length (job->priv->LY_LIST) && 
+	if (job->priv->LY_LIST && (g_list_length (job->priv->LY_LIST) > 1) && 
 			!((g_list_length (job->priv->LY_LIST) == 1) &&
 				(((gdouble *) g_list_nth_data (job->priv->LY_LIST, 0)) [0] == 1.) &&
 				(((gdouble *) g_list_nth_data (job->priv->LY_LIST, 0)) [1] == 0.) &&
@@ -363,7 +406,6 @@ gnome_print_job_setup_context (GnomePrintJob *job, GnomePrintContext *pc)
 				(((gdouble *) g_list_nth_data (job->priv->LY_LIST, 0)) [3] == 1.) &&
 				(((gdouble *) g_list_nth_data (job->priv->LY_LIST, 0)) [4] == 0.) &&
 				(((gdouble *) g_list_nth_data (job->priv->LY_LIST, 0)) [5] == 0.))) {
-		GnomePrintFilter *multipage;
 		GValue vd = {0,};
 		GValueArray *va = g_value_array_new (0);
 		guint i, npages;
@@ -381,16 +423,21 @@ gnome_print_job_setup_context (GnomePrintJob *job, GnomePrintContext *pc)
 			}
 		}
 		g_value_unset (&vd);
-		multipage = gnome_print_filter_new_from_module_name ("multipage",
+		f = gnome_print_filter_new_from_module_name ("multipage",
 				"affines", va, NULL);
 		g_value_array_free (va);
-		if (multipage) {
+		if (f) {
+			g_object_get (G_OBJECT (pc), "filter", &filter, NULL);
 			if (filter)
-				gnome_print_filter_append_predecessor (filter, multipage);
-			g_object_set (G_OBJECT (pc), "filter", multipage, NULL);
-			g_object_unref (G_OBJECT (multipage));
+				gnome_print_filter_append_predecessor (filter, f);
+			g_object_set (G_OBJECT (pc), "filter", f, NULL);
+			g_object_unref (G_OBJECT (f));
 		}
 	}
+
+	g_object_get (G_OBJECT (pc), "filter", &filter, NULL);
+	if (filter)
+		gnome_print_filter_reset (filter);
 }
 
 /**
@@ -421,6 +468,7 @@ gnome_print_job_get_pages (GnomePrintJob *job)
 	pc = g_object_new (GNOME_TYPE_PRINT_META, NULL);
 	gnome_print_job_setup_context (job, pc);
 	gnome_print_meta_render (GNOME_PRINT_META (job->meta), pc);
+	gnome_print_context_close (pc);
 	n = gnome_print_meta_get_pages (GNOME_PRINT_META (pc));
 	g_object_unref (G_OBJECT (pc));
 
@@ -581,8 +629,6 @@ gnome_print_job_print (GnomePrintJob *job)
 	gint lpages, copies, nstacks, npages, nsheets;
 	gboolean collate, hwcopies;
 	gint stack;
-	const guchar *buf;
-	gint blen;
 	gint ret;
 
 	g_return_val_if_fail (job != NULL, GNOME_PRINT_ERROR_UNKNOWN);
@@ -591,7 +637,7 @@ gnome_print_job_print (GnomePrintJob *job)
 
 	if (job->input_file) {
 		GnomePrintTransport *transport = gnome_print_transport_new (job->config);
-		return gnome_print_transport_print_file (transport, job->input_file);
+		return gnome_print_transport_print_file (transport, (const guchar *) job->input_file);
 	}
 
 	if (!job->priv->closed) {
@@ -636,7 +682,11 @@ gnome_print_job_print (GnomePrintJob *job)
 		nsheets = copies;
 	}
 
-	npages = (lpages + job->priv->num_affines - 1) / job->priv->num_affines;
+	if (job->priv->num_affines >= 1) {
+		npages = (lpages + job->priv->num_affines - 1) / job->priv->num_affines;
+	} else {
+		npages = 0;
+	}
 	for (stack = 0; stack < nstacks; stack++) {
 		gint page;
 		for (page = 0; page < npages; page++) {
@@ -650,17 +700,19 @@ gnome_print_job_print (GnomePrintJob *job)
 							ctx, i, TRUE);
 					g_return_val_if_fail (ret == GNOME_PRINT_OK, ret);
 				}
-				/* Flush physical page */
-				if (filter)
-					gnome_print_filter_flush (filter);
 			}
 		}
 		if (stack + 1 < nstacks) {
+			if (filter)
+				gnome_print_filter_flush (filter);
 			ret = gnome_print_end_doc (ctx);
 			g_return_val_if_fail (ret == GNOME_PRINT_OK, ret);
 		}
 	}
 
+	if (NULL != filter)
+	    g_object_unref (G_OBJECT (filter));
+
 	ret = gnome_print_context_close (ctx);
 	g_object_unref (G_OBJECT (ctx));
 
@@ -708,8 +760,10 @@ gnome_print_job_render_page (GnomePrintJob *job, GnomePrintContext *ctx, gint pa
 	pc = g_object_new (GNOME_TYPE_PRINT_META, NULL);
 	gnome_print_job_setup_context (job, pc);
 	gnome_print_meta_render (GNOME_PRINT_META (job->meta), pc);
+	gnome_print_context_close (pc);
 	gnome_print_meta_render_page (GNOME_PRINT_META (pc), ctx, page, pageops);
 	g_object_unref (G_OBJECT (pc));
+	return GNOME_PRINT_OK;
 }
 
 /* We need:
